#include "Game2048.hpp"
#include <iostream>
#include <cstdlib>
#include <ctime>
#include <algorithm>

// The creater for the 2048 game class - initializes the board, seeds the random number generator and generate initial cells.
Game2048::Game2048(){
    board = std::vector<std::vector<int>>(4, std::vector<int>(4, 0));
    srand(time(0));
    generateNewTile();

}

// The main loop below runs the game - start the game loop, print the board, read user input, process the move, generate new cell and end the game (when the while loop ends) 

void Game2048::start(){
    while (!isGameOver()){
        printBoard();
        char move;
        std::cin >> move;
        switch (move) {
            case 'j' : moveUp(); break;
            case 'n' : moveDown(); break;
            case 'b' : moveLeft(); break;
            case 'm' : moveRight(); break;
            default: continue;
        }
        generateNewTile();
    }
    std::cout << "Game Over!" << std::endl;
}

// The next function is responsible for printing the current state of the game - cycles through rows, cycles through cells in a row, prints cell value and prints newline after each row.

void Game2048::printBoard() {
    for (const auto& row : board) {
        for (const auto& cell : row){
            std::cout << cell << "\t";

        }
        std::cout << std::endl;
    }
}

// The next 4 functions handle the logic for moving the tiles to the left, right, up and down - intializes a moved (a boolean flag) to false, loops through rows, merge row tiles, shift nonzero cells and return moved.
// moveLeft funtion

bool Game2048::moveLeft(){
    bool moved = false;
    for (auto& row : board) {
        merge(row);
        for(int i = 0; i < 4; ++i){
            if (row[i] == 0){
                for (int j = i + 1; j < 4; ++j) {
                    if(row[j] != 0) {
                        row[i] = row[j];
                        row[j] = 0;
                        moved = true;
                        break;
                    }

                }
            }
        }
    }
    return moved;
}

// moveRight function

bool Game2048::moveRight() {
    bool moved = false;
    for (auto& row : board) {
        std::reverse(row.begin(), row.end());
        merge (row);
        for (int i = 0; i < 4; ++i){
            if (row[i] == 0) {
                for (int j = i + 1; j < 4; ++j) {
                    if (row[j] != 0) {
                        row[i] = row[j];
                        row[j] = 0;
                        moved = true;
                        break;
                    }
                }
            }
        }
        std::reverse(row.begin(), row.end());
    }
    return moved;

}

// moveUp function

bool Game2048::moveUp() {
    bool moved = false;
    for (int col = 0; col < 4; ++col) {
        std::vector<int> column(4);
        for (int row = 0; row < 4; ++row) {
            column[row] = board[row][col];
        }
        merge (column);
        for (int i = 0; i < 4; ++i) {
            if (column[i] == 0) {
                for(int j = i + 1; j < 4; ++j) {
                    if (column[j] != 0){
                        column[i] = column[j];
                        column[j] = 0;
                        moved = true;
                        break;
                    }
                }
            }
        }
        for (int row = 0; row < 4; ++row) {
            board[row][col] = column[row];
        }
    }
    return moved;
}

// moveDown function 

bool Game2048::moveDown() {
    bool moved = false;
    for (int col = 0; col < 4; ++col) {
        std::vector<int> column(4);
        for (int row = 0; row < 4; ++row) {
            column[row] = board[row][col];
        }
        std::reverse(column.begin(), column.end());
        merge(column);
        for (int i = 0; i < 4; ++i) {
            if(column[i] == 0) {
                for (int j = i + 1; j < 4; ++j) {
                    if (column[j] != 0) {
                        column[i] = column[j];
                        column[j] = 0;
                        moved = true;
                        break;
                    }
                }

            }
        }
        std::reverse(column.begin(), column.end());
        for (int row  = 0; row < 4; ++row) {
            board[row][col] = column[row];
        }
    }
    return moved;
}

// the next function checks if the game is over by determining if there are no more empty cell - loops through rows, loops through cells in each row and returns true if no empty cell

bool Game2048::isGameOver() {
    for (const auto& row : board) {
        for (const auto& cell : row) {
            if(cell == 0) return false;
        }
    }
    return true;
}

// the next function is responsible for adding a new cell to the game board in a random empty postion - declare row and column variable, find a random empty position and place a new cell.

void Game2048::generateNewTile() {
    int row, col;
    do {
        row = rand() % 4;
        col = rand() % 4;
    } 
    while (board[row][col] != 0);
    board[row][col] = (rand() % 2 + 1) * 2;
}

// the final function is responsible for the merging of the cells when making contact

void Game2048::merge(std::vector<int>& row) {
    for (int i = 0; i < 3; ++i) {
        if (row[i] != 0 && row[i] == row[i + 1]) {
            row[i] *= 2;
            row[i + 1] = 0; 
            ++i;
        }
    }
}